import asyncio
import json
from asyncio import Task

import aiohttp
import httpx
from fastapi import HTTPException
from starlette import status

from database.models.nodes import Node
from database.schemas.cams import CamOut


class NodeConnection:

    def __init__(self, node: Node):
        self._node = node
        self._connect_task = None
        self._base_url = f"{'https' if self._node.ssl else 'http'}://" \
                         f"{self._node.host}:{self._node.port}"

    @property
    def id(self) -> int:
        return int(self._node.id)

    @property
    def connected(self) -> bool:
        return not isinstance(self._connect_task, Task)

    async def _connect_node(self):
        url = f"{self._base_url}/set/{int(self._node.id)}"
        while True:
            try:
                async with aiohttp.ClientSession() as session:
                    response = await session.post(url)
                    if response.status != status.HTTP_200_OK:
                        raise HTTPException(status_code=response.status)
                break
            except (httpx.ConnectTimeout, httpx.ConnectError) as e:
                print(f"Failed to connect to node {self.id}: {e}")
                await asyncio.sleep(5)
            except asyncio.CancelledError:
                print(f"Connection to node {self.id} cancelled.")
                return -1
        self._connect_task = None

    async def connect(self):
        self._connect_task = asyncio.create_task(self._connect_node())

    async def check(self) -> bool:
        url = f"{self._base_url}/"
        async with aiohttp.ClientSession() as session:
            response = await session.post(url)
            if response.status != status.HTTP_200_OK:
                return False

            try:
                parsed = json.loads(await response.read())
                return parsed['id'] == int(self._node.id)
            finally:
                return False

    async def get_active_cam(self) -> CamOut:
        url = f"{self._base_url}/active_cam"
        async with aiohttp.ClientSession() as session:
            response = await session.get(url)
            if response.status != status.HTTP_200_OK:
                raise HTTPException(status_code=response.status)

            parsed = json.loads(await response.read())
            return CamOut(**parsed)

    async def choose_camera(self, cam_id: int) -> CamOut:
        url = f"{self._base_url}/choose_camera/{cam_id}"
        async with aiohttp.ClientSession() as session:
            response = await session.get(url)
            if response.status != status.HTTP_200_OK:
                raise HTTPException(status_code=response.status)

            parsed = json.loads(await response.read())
            return CamOut(**parsed)

    async def get_image(self) -> bytes:
        url = f"{self._base_url}/snapshot"
        async with aiohttp.ClientSession() as session:
            response = await session.get(url)
            if response.status != status.HTTP_200_OK:
                raise HTTPException(status_code=response.status)

            return await response.read()

    async def a_cancel(self):
        if isinstance(self._connect_task, Task):
            self._connect_task.cancel()
            try:
                await self._connect_task
            except asyncio.CancelledError:
                pass

    def cancel(self):
        if isinstance(self._connect_task, Task):
            self._connect_task.cancel()

    def __del__(self):
        self.cancel()
