import base64
import hashlib
from datetime import datetime
from typing import Set, Dict

from lxml import etree

import config


class NonceCache:
    def __init__(self, size: int, cleanup_threshold: int):
        self.size = size
        self.cleanup_threshold = cleanup_threshold

        self._nonce_set: Set[bytes] = set()
        self._timestamp_map: Dict[datetime, bytes] = dict()
        self._min_timestamp = None

    def cleanup(self):
        if len(self._nonce_set) >= self.cleanup_threshold:
            while len(self._nonce_set) >= self.size:
                key = min(self._timestamp_map)
                nonce = self._timestamp_map.pop(key)
                self._nonce_set.discard(nonce)
            self._min_timestamp = min(self._timestamp_map, default=None)

    def check(self, nonce: bytes, timestamp: datetime) -> bool:
        if (nonce in self._nonce_set or
                (self._min_timestamp is not None and
                 timestamp < self._min_timestamp)):
            return False
        else:
            self._nonce_set.add(nonce)
            self._timestamp_map[timestamp] = nonce
            self.cleanup()
            return True


def make_digest(password: bytes | str, nonce: bytes, timestamp: str):
    if isinstance(password, str):
        password = password.encode('utf-8')
    digest = base64.b64encode(
        hashlib.sha1(
            nonce + timestamp.encode('utf-8') + password
        ).digest()
    ).decode('ascii')
    return digest


nonce_cache = NonceCache(int(5e6), int(6e6))


def check_password(xml_data: bytes) -> bool:
    root = etree.fromstring(xml_data)

    # Search for Security block
    nsmap = {
        'wsse': "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd",
        'wsu': "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd"
    }
    security_elem = root.find(".//wsse:Security", nsmap)
    if security_elem is None:
        return False

    username_token_elem = security_elem.find(".//wsse:UsernameToken", nsmap)
    if username_token_elem is None:
        return False

    username_elem = username_token_elem.find(".//wsse:Username", nsmap)
    if username_elem is None or username_elem.text != config.credentials['login']:
        return False

    digest_elem = username_token_elem.find(".//wsse:Password", nsmap)
    nonce_elem = username_token_elem.find(".//wsse:Nonce", nsmap)
    created_elem = username_token_elem.find(".//wsu:Created", nsmap)

    if digest_elem is None or nonce_elem is None or created_elem is None:
        return False

    try:
        datetime_obj = datetime.fromisoformat(created_elem.text)
        nonce = base64.b64decode(nonce_elem.text)
        if not nonce_cache.check(nonce, datetime_obj):
            return False

        return (digest_elem.text ==
                make_digest(config.credentials['password'], nonce, created_elem.text))

    except ValueError:
        return False
