import asyncio
from asyncio import Task
from typing import Dict, List

import httpx
from fastapi import HTTPException
from onvif import ONVIFError
from starlette import status

from database import db_connection as db
from database.repositories.cams import CamRepository
from .camera import Camera


class HotCams:
    class CamConnection:
        def __init__(self):
            self._connect_task = None
            self.camera: Camera | None = None

        async def _create_camera(self, _cam_id, **kwargs):
            while True:
                try:
                    cam = Camera(_cam_id, **kwargs)
                    await cam.update()
                    print(f"Connection to camera {kwargs['name']} established.")
                    break
                except (ONVIFError, httpx.ConnectTimeout, httpx.ConnectError) as e:
                    if 'name' in kwargs:
                        print(f"Failed to connect to camera {kwargs['name']}: {e}")
                    await asyncio.sleep(5)
                except asyncio.CancelledError:
                    print(f"Connection to camera {kwargs['name']} cancelled.")
                    return -1
            self.camera = cam
            self._connect_task = None
            return _cam_id

        async def try_connect(self, cam_id: int):
            async with db.get_session() as session:
                cam_repo = CamRepository(session)
                cam_db_obj = await cam_repo.get_or_none(cam_id)
                if cam_db_obj is None:
                    raise HTTPException(
                        status_code=status.HTTP_404_NOT_FOUND
                    )

                params = cam_db_obj.to_dict()
                del params['id']
                self._connect_task = asyncio.create_task(self._create_camera(cam_id, **params))

        async def a_cancel(self):
            if isinstance(self._connect_task, Task):
                self._connect_task.cancel()
                try:
                    await self._connect_task
                except asyncio.CancelledError:
                    pass

        def cancel(self):
            if isinstance(self._connect_task, Task):
                self._connect_task.cancel()

        def __del__(self):
            self.cancel()

    def __init__(self):
        self._connections: Dict[int, HotCams.CamConnection] = {}
        self._current_camera: Camera | None = None

    @property
    def current_camera(self) -> Camera:
        return self._current_camera

    async def _connect_camera(self, cam_id: int) -> CamConnection:
        connection = self._connections.get(cam_id, None)
        if connection is None:
            connection = HotCams.CamConnection()
            await connection.try_connect(cam_id)
            self._connections[cam_id] = connection
        return connection

    async def connect_camera(self, cam_id: int):
        await self._connect_camera(cam_id)

    async def connect_cams_list(self, cam_ids: List[int]):
        for cam_id in cam_ids:
            await self._connect_camera(cam_id)

    async def try_connect_cams_list(self, cam_ids: List[int]):
        for cam_id in cam_ids:
            asyncio.create_task(self._connect_camera(cam_id))

    async def disconnect_camera(self, cam_id: int):
        connection = self._connections.get(cam_id, None)
        if connection is None:
            return

        await connection.a_cancel()

    async def get_cam_by_id(self, cam_id: int):
        connection = await self._connect_camera(cam_id)
        cam = connection.camera
        if cam is None:
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE
            )
        return cam

    async def switch_camera(self, cam_id: int):
        self._current_camera = await self.get_cam_by_id(cam_id)
        return self._current_camera

    def __del__(self):
        for _, connection in self._connections:
            if connection is not None:
                connection.cancel()
