import asyncio
from lxml import etree
from fastapi import APIRouter
from starlette.responses import FileResponse

from node_settings.node_settings import current_node

from .proxy_utils import onvif_proxy, prevent_self_closing_tags, token_spoofing

import config

onvif_router = APIRouter()


# noinspection HttpUrlsUsage
@onvif_router.post("/device_service")
@onvif_proxy()
async def device_service_proxy(response_body):
    # Parsing XML response
    root = etree.fromstring(response_body)

    # Searching for blocks to proxy
    for service in config.services_to_proxy:
        ptz_elem = root.find(f".//{{http://www.onvif.org/ver10/schema}}{service}")
        if ptz_elem is not None:
            # Replacing a link in a PTZ block
            xaddr_elem = ptz_elem.find(".//{http://www.onvif.org/ver10/schema}XAddr")
            if xaddr_elem is not None:
                xaddr_elem.text = f"http://{current_node.host}:{current_node.port}/onvif/{service}"

    # Converting XML back to string
    prevent_self_closing_tags(root)
    return etree.tostring(root, encoding='utf-8')


# noinspection HttpUrlsUsage
@onvif_router.post("/Media")
@onvif_proxy()
async def media_proxy(response_body):
    ns = {
        'trt': "http://www.onvif.org/ver10/media/wsdl",
        'tt': "http://www.onvif.org/ver10/schema"
    }

    # Parsing XML response
    root = etree.fromstring(response_body)

    # Substitution of stream link
    stream_elem = root.find(".//trt:GetStreamUriResponse", ns)
    if stream_elem is not None:
        media_elem = stream_elem.find(".//trt:MediaUri", ns)
        if media_elem is not None:
            uri_elem = stream_elem.find(".//tt:Uri", ns)
            if uri_elem is not None:
                uri_elem.text = f"rtsp://{current_node.host}:{current_node.rtsp_port}/stream"

    # Replacing snapshot link
    stream_elem = root.find(".//trt:GetSnapshotUriResponse", ns)
    if stream_elem is not None:
        media_elem = stream_elem.find(".//trt:MediaUri", ns)
        if media_elem is not None:
            uri_elem = stream_elem.find(".//tt:Uri", ns)
            if uri_elem is not None:
                uri_elem.text = f"http://{current_node.host}:{current_node.port}/onvif/snapshot"

    # Deleting all profiles except the first one
    profiles_resp_elem = root.find(".//trt:GetProfilesResponse", ns)
    if profiles_resp_elem is not None:
        profiles_elem = profiles_resp_elem.findall(".//trt:Profiles", ns)
        for profile in profiles_elem[1:]:
            profiles_resp_elem.remove(profile)

    # Converting XML back to string
    prevent_self_closing_tags(root)
    return etree.tostring(root, encoding='utf-8')


@onvif_router.post("/PTZ")
@onvif_proxy(request_handler=token_spoofing)
async def ptz_proxy(response_body):
    return response_body


@onvif_router.get("/snapshot")
async def get_image():
    # Return a file response with the specified file path
    return FileResponse("onvif_proxy/snapshot.jpg", media_type="image/jpeg")


@onvif_router.post("/{path:path}")
@onvif_proxy()
async def catch_all_proxy(response_body):
    print("Oops! Used default proxy:")
    return response_body
