import aiohttp

from lxml import etree
from fastapi import Request, HTTPException, status
from fastapi.responses import Response
from typing import Callable, Awaitable

from database.models.nodes import Node
from node_settings.node_settings import current_node
from node_switch.switch_utils import get_current_cam
from .auth_utils import check_password


def prevent_self_closing_tags(xml_etree):
    for node in xml_etree.iter():
        if node.text is None:
            node.text = ''


def print_xml(xml_data):
    xml_parsed = etree.fromstring(xml_data)
    xml_pretty = etree.tostring(xml_parsed, pretty_print=True).decode('utf-8')
    print(xml_pretty)


# noinspection HttpUrlsUsage
def security_spoofing(xml_data: bytes):
    camera = get_current_cam()

    # Parsing XML response
    root = etree.fromstring(xml_data)

    # Search for Security block
    ns = {
        'ns1': "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd",
        'ns2': "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd"
    }
    security_elem = root.find(".//ns1:Security", ns)
    if security_elem is not None:
        security_elem.getparent().replace(security_elem, camera.make_security_etree())

    # Converting XML back to string
    prevent_self_closing_tags(root)
    return etree.tostring(root, encoding='utf-8')


# noinspection HttpUrlsUsage
async def token_spoofing(xml_data: bytes):
    camera = get_current_cam()

    # Parsing XML response
    root = etree.fromstring(xml_data)

    # Search for a token
    ns = {
        'soap-env': "http://www.w3.org/2003/05/soap-envelope",
        'tptz10': "http://www.onvif.org/ver10/ptz/wsdl",
        'tptz20': "http://www.onvif.org/ver20/ptz/wsdl"
    }
    body_elem = root.find(".//soap-env:Body", ns)
    if body_elem is not None:
        tokens_elem = body_elem.findall(".//tptz10:ProfileToken", ns)
        for token in tokens_elem:
            token.text = camera.profiles[0].token

        tokens_elem = body_elem.findall(".//tptz20:ProfileToken", ns)
        for token in tokens_elem:
            token.text = camera.profiles[0].token

    # Converting XML back to string
    prevent_self_closing_tags(root)
    return etree.tostring(root, encoding='utf-8')


# noinspection HttpUrlsUsage
def onvif_proxy(request_handler: Callable[[bytes], Awaitable] = None,
                print_request=False,
                print_response=False,
                print_raw_request=False,
                print_raw_response=False):
    def wrap(xml_handler):
        async def inner(request: Request):
            if not isinstance(current_node, Node):
                raise HTTPException(
                    status_code=status.HTTP_503_SERVICE_UNAVAILABLE
                )

            camera = get_current_cam()

            # Receiving the request body
            request_body = (await request.body()).replace(b'\r', b'').replace(b'\n', b'')
            request_headers = {key.lower(): str(value) for key, value in request.headers.items()}

            # Setting headers
            if 'host' in request_headers:
                request_headers['host'] = f'{camera.host}:{camera.port}'

            request_url = request.url.path
            del request

            if request_handler is not None:
                request_body = await request_handler(request_body)

            if print_raw_request:
                print(f"Received SOAP request ({request_url}):")
                print_xml(request_body)

            if not check_password(request_body):
                return Response(content=unauthorized_xml, status_code=401)

            # Substitution of authorization data
            request_body = security_spoofing(request_body)

            if print_request:
                print(f"Sending SOAP request ({request_url}):")
                print_xml(request_body)

            if 'content-length' in request_headers:
                request_headers['content-length'] = str(len(request_body))

            async with aiohttp.ClientSession() as session:
                # noinspection HttpUrlsUsage
                async with session.post(f"http://{camera.host}:{camera.port}{request_url}", data=request_body,
                                        headers=request_headers) as response:
                    response_body = (await response.read()).replace(b'\r', b'').replace(b'\n', b'')
                    response_headers = {key.lower(): str(value) for key, value in response.headers.items()}
                    response_status = response.status

            if print_raw_response:
                print(f"Received SOAP response ({request_url}):")
                print_xml(response_body)

            # Calling the response handler
            response_body = await xml_handler(response_body)

            if print_response:
                print(f"Sending SOAP response ({request_url}):")
                print_xml(response_body)

            # Setting headers
            if 'content-length' in response_headers:
                response_headers['content-length'] = str(len(response_body))
            if 'connection' in response_headers:
                del response_headers['connection']
            if 'server' in response_headers:
                del response_headers['server']

            return Response(
                status_code=response_status,
                content=response_body,
                headers=response_headers
            )

        return inner

    return wrap


# noinspection HttpUrlsUsage
def make_unauthorized_xml():
    nsmap = {
        "SOAP-ENV": "http://www.w3.org/2003/05/soap-envelope",
        "SOAP-ENC": "http://www.w3.org/2003/05/soap-encoding",
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "xsd": "http://www.w3.org/2001/XMLSchema",
        "xs": "http://www.w3.org/2001/XMLSchema",
        "wsa5": "http://schemas.xmlsoap.org/ws/2004/08/addressing",
        "c14n": "http://www.w3.org/2001/10/xml-exc-c14n#",
        "wsu": "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd",
        "xenc": "http://www.w3.org/2001/04/xmlenc#",
        "ds": "http://www.w3.org/2000/09/xmldsig#",
        "wsse": "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd",
        "wsa": "http://schemas.xmlsoap.org/ws/2004/08/addressing",
        "xmime": "http://tempuri.org/xmime.xsd",
        "xop": "http://www.w3.org/2004/08/xop/include",
        "wsrf-bf": "http://docs.oasis-open.org/wsrf/bf-2",
        "d": "http://schemas.xmlsoap.org/ws/2005/04/discovery",
        "tt": "http://www.onvif.org/ver10/schema",
        "wstop": "http://docs.oasis-open.org/wsn/t-1",
        "wsrf-r": "http://docs.oasis-open.org/wsrf/r-2",
        "ns1": "http://www.onvif.org/ver10/network/wsdl/RemoteDiscoveryBinding",
        "ns10": "http://www.onvif.org/ver10/events/wsdl/PausableSubscriptionManagerBinding",
        "ns11": "http://www.onvif.org/ver20/analytics/wsdl/RuleEngineBinding",
        "ns12": "http://www.onvif.org/ver20/analytics/wsdl/AnalyticsEngineBinding",
        "tan": "http://www.onvif.org/ver20/analytics/wsdl",
        "ns2": "http://www.onvif.org/ver10/network/wsdl/DiscoveryLookupBinding",
        "dn": "http://www.onvif.org/ver10/network/wsdl",
        "ns3": "http://www.onvif.org/ver10/events/wsdl/PullPointSubscriptionBinding",
        "ns4": "http://www.onvif.org/ver10/events/wsdl/EventBinding",
        "tev": "http://www.onvif.org/ver10/events/wsdl",
        "ns5": "http://www.onvif.org/ver10/events/wsdl/SubscriptionManagerBinding",
        "ns6": "http://www.onvif.org/ver10/events/wsdl/NotificationProducerBinding",
        "ns7": "http://www.onvif.org/ver10/events/wsdl/NotificationConsumerBinding",
        "ns8": "http://www.onvif.org/ver10/events/wsdl/PullPointBinding",
        "ns9": "http://www.onvif.org/ver10/events/wsdl/CreatePullPointBinding",
        "wsnt": "http://docs.oasis-open.org/wsn/b-2",
        "tds": "http://www.onvif.org/ver10/device/wsdl",
        "timg10": "http://www.onvif.org/ver10/imaging/wsdl",
        "timg20": "http://www.onvif.org/ver20/imaging/wsdl",
        "tptz10": "http://www.onvif.org/ver10/ptz/wsdl",
        "tptz20": "http://www.onvif.org/ver20/ptz/wsdl",
        "tmd": "http://www.onvif.org/ver10/deviceIO/wsdl",
        "trt": "http://www.onvif.org/ver10/media/wsdl",
        "ter": "http://www.onvif.org/ver10/error",
        "tns1": "http://www.onvif.org/ver10/topics",
        "tnsn": "http://www.eventextension.com/2011/event/topics",
        "extwsd": "http://www.onvifext.com/onvif/ext/ver10/wsdl",
        "extxsd": "http://www.onvifext.com/onvif/ext/ver10/schema"
    }

    root = etree.Element("{SOAP-ENV}Envelope", nsmap=nsmap)
    body = etree.SubElement(root, "{SOAP-ENV}Body", nsmap=nsmap)
    fault = etree.SubElement(body, "{SOAP-ENV}Fault", nsmap=nsmap)
    code = etree.SubElement(fault, "{SOAP-ENV}Code", nsmap=nsmap)
    value = etree.SubElement(code, "{SOAP-ENV}Value", nsmap=nsmap)
    value.text = "SOAP-ENV:Sender"
    reason = etree.SubElement(fault, "{SOAP-ENV}Reason", nsmap=nsmap)
    text = etree.SubElement(reason, "{SOAP-ENV}Text", nsmap=nsmap)
    text.set("{http://www.w3.org/XML/1998/namespace}lang", "en")
    text.text = "HTTP Error: 401 Unauthorized"

    return etree.tostring(root, encoding="utf-8")


unauthorized_xml = make_unauthorized_xml()
