from typing import List

from fastapi import APIRouter, HTTPException, status, Depends

from database import db_connection as db
from database.repositories.nodes import NodeRepository
from database.schemas.nodes import NodeCreate, NodeFull, NodeUpdate

db_nodes_router = APIRouter()


@db_nodes_router.get("/list", response_model=List[NodeFull])
async def list_nodes(
        node_repo: NodeRepository = Depends(db.get_repository(NodeRepository))
):
    nodes = await node_repo.list_all()
    if nodes is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND
        )
    return nodes


@db_nodes_router.get("/{node_id}", response_model=NodeFull)
async def get_node(
        node_id: int,
        node_repo: NodeRepository = Depends(db.get_repository(NodeRepository))
):
    node = await node_repo.get(node_id)
    return node


@db_nodes_router.post("", response_model=NodeFull)
async def create_node(
    node: NodeCreate,
    node_repo: NodeRepository = Depends(db.get_repository(NodeRepository))
):
    new_node = await node_repo.create(node)
    if new_node is None:
        raise HTTPException(
            status_code=status.HTTP_304_NOT_MODIFIED
        )
    return new_node


@db_nodes_router.patch("/{node_id}", response_model=NodeFull)
async def update_node(
    node_id: int,
    node: NodeUpdate,
    node_repo: NodeRepository = Depends(db.get_repository(NodeRepository))
):
    updated_node = await node_repo.update(node_id, node)
    if updated_node is None:
        raise HTTPException(
            status_code=status.HTTP_304_NOT_MODIFIED
        )

    return updated_node


@db_nodes_router.delete("/{node_id}", response_model=NodeFull)
async def delete_node(
        node_id: int,
        node_repo: NodeRepository = Depends(db.get_repository(NodeRepository))
):
    node = await node_repo.delete(node_id)
    if node is None:
        raise HTTPException(
            status_code=status.HTTP_304_NOT_MODIFIED
        )
    return node
