from typing import Any, Type, AsyncGenerator

from fastapi import Depends
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.orm import sessionmaker

import config
from database.models.base import Base
from database.repositories.base import Repository


class AsyncDBConnection:
    class WithSession:
        def __init__(self, session_maker):
            self._maker = session_maker
            self._session = None

        async def __aenter__(self):
            self._session = self._maker()
            return self._session

        async def __aexit__(self, exc_type, exc, tb):
            if isinstance(self._session, AsyncSession):
                await self._session.close()

    def __init__(self, url: str):
        self._engine = create_async_engine(url, echo=False)
        # noinspection PyTypeChecker
        self._session_maker = \
            sessionmaker(
                self._engine,
                expire_on_commit=False,
                class_=AsyncSession
            )

    async def create_models(self):
        async with self._engine.begin() as connection:
            await connection.run_sync(Base.metadata.create_all)

    async def _get_db(self) -> AsyncGenerator:
        db = None
        try:
            db = self._session_maker()
            yield db
        finally:
            if isinstance(db, AsyncSession):
                await db.close()

    def get_session(self):
        return self.WithSession(self._session_maker)

    def get_repository(
            self,
            repository: Type[Repository],
    ) -> Any:
        def repository_getter(db: AsyncSession = Depends(self._get_db)) -> Repository:
            return repository(db)

        return repository_getter


def make_db_url() -> str:
    db = config.database
    return (f"{db['driver']}://"
            f"{db['user']}:{db['password']}@"
            f"{db['host']}:{db['port']}/{db['database']}")
