from typing import List

from fastapi import HTTPException
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from starlette import status

from database.models.auth import User
from database.models.cams import Group, UserInGroup, Cam, CamInGroup
from database.models.nodes import Node
from database.repositories.base import Repository
from database.schemas.cam_groups import GroupBase


# noinspection DuplicatedCode
class GroupRepository(Repository):
    def __init__(self, db_session: AsyncSession):
        super().__init__(db_session)

    async def list_all(self) -> List[Group]:
        query = select(Group)
        group_db_obj = await self._session.execute(query)
        return [group for group, in group_db_obj]

    async def get_or_none(self, group_id: int) -> Group | None:
        group_db_obj = await self._session.get(Group, group_id)
        return group_db_obj

    async def get(self, group_id: int) -> Group:
        group_db_obj = await self.get_or_none(group_id)
        if group_db_obj is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Group with id={group_id} not found"
            )
        return group_db_obj

    async def create(self, group: GroupBase) -> Group:
        group_db_obj = Group(**group.model_dump(exclude_unset=True))
        self._session.add(group_db_obj)
        await self._session.commit()
        await self._session.refresh(group_db_obj)
        return group_db_obj

    async def update(self, group_id: int, group: GroupBase) -> Group | None:
        group_db_obj = await self._session.get(Group, group_id)
        for key, value in group.model_dump(exclude_unset=True).items():
            setattr(group_db_obj, key, value)

        await self._session.commit()
        await self._session.refresh(group_db_obj)
        return group_db_obj

    async def delete(self, group_id: int) -> Group | None:
        group_db_obj = await self._session.get(Group, group_id)
        await self._session.delete(group_db_obj)
        await self._session.commit()
        return group_db_obj

    async def list_cams_in_group(self, group: Group) -> List[Cam]:
        query = select(Cam).join(CamInGroup, CamInGroup.cam_id == Cam.id)
        query = query.filter(CamInGroup.group_id == group.id)
        cams_db_obj = await self._session.execute(query)
        return [cam for cam, in cams_db_obj]

    async def get_user_groups(self, user: User) -> List[Group]:
        query = select(Group).join(UserInGroup, UserInGroup.group_id == Group.id)
        query = query.filter(UserInGroup.user_id == user.id)
        groups_db_obj = await self._session.execute(query)
        return [group for group, in groups_db_obj]

    async def add_user_group(self, user: User, group: Group) -> UserInGroup:
        user_group_db_obj = UserInGroup(user_id=user.id, group_id=group.id)
        self._session.add(user_group_db_obj)
        await self._session.commit()
        await self._session.refresh(user_group_db_obj)
        return user_group_db_obj

    async def delete_user_group(self, user: User, group: Group) -> UserInGroup | None:
        user_group_db_obj = await self._session.get(UserInGroup, [user.id, group.id])
        await self._session.delete(user_group_db_obj)
        await self._session.commit()
        return user_group_db_obj
