from typing import List

from fastapi import HTTPException
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from starlette import status

from database.models.cams import Group, Cam, CamInGroup
from database.models.nodes import Node
from database.repositories.base import Repository
from database.schemas.cams import CamCreate, CamUpdate


# noinspection DuplicatedCode
class CamRepository(Repository):
    def __init__(self, db_session: AsyncSession):
        super().__init__(db_session)

    async def list_all(self) -> List[Cam]:
        query = select(Cam)
        cam_db_obj = await self._session.execute(query)
        return [cam for cam, in cam_db_obj]

    async def get_or_none(self, cam_id: int) -> Cam | None:
        cam_db_obj = await self._session.get(Cam, cam_id)
        return cam_db_obj

    async def get(self, cam_id: int) -> Cam:
        cam_db_obj = await self.get_or_none(cam_id)
        if cam_db_obj is None:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Cam with id={cam_id} not found"
            )
        return cam_db_obj

    async def create(self, cam: CamCreate) -> Cam:
        cam_db_obj = Cam(**cam.model_dump(exclude_unset=True))
        self._session.add(cam_db_obj)
        await self._session.commit()
        await self._session.refresh(cam_db_obj)
        return cam_db_obj

    async def update(self, cam_id: int, cam: CamUpdate) -> Cam | None:
        cam_db_obj = await self._session.get(Cam, cam_id)
        for key, value in cam.model_dump(exclude_unset=True).items():
            setattr(cam_db_obj, key, value)

        await self._session.commit()
        await self._session.refresh(cam_db_obj)
        return cam_db_obj

    async def delete(self, cam_id: int) -> Cam | None:
        cam_db_obj = await self._session.get(Cam, cam_id)
        await self._session.delete(cam_db_obj)
        await self._session.commit()
        return cam_db_obj

    async def get_cam_groups(self, cam: Cam) -> List[Group]:
        query = select(Group).join(CamInGroup, CamInGroup.cam_id == Group.id)
        query = query.filter(CamInGroup.cam_id == cam.id)
        groups_db_obj = await self._session.execute(query)
        return [group for group, in groups_db_obj]

    async def add_cam_group(self, group: Group, cam: Cam, node: Node = None) -> CamInGroup:
        cam_group_db_obj = CamInGroup(cam_id=cam.id, group_id=group.id, node_id=node.id)
        self._session.add(cam_group_db_obj)
        await self._session.commit()
        await self._session.refresh(cam_group_db_obj)
        return cam_group_db_obj

    async def delete_cam_group(self, cam: Cam, group: Group) -> CamInGroup | None:
        cam_group_db_obj = await self._session.get(CamInGroup, [cam.id, group.id])
        await self._session.delete(cam_group_db_obj)
        await self._session.commit()
        return cam_group_db_obj
